/*
 * Copyright (C) 2014 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <sys/file.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "libtp.h"

static void wanting(char *prefix)
{
	pid_t pid;

	pid = getpid();
	if (!prefix)
		printf("pid %d is wanting ...\n", pid);
	else
		printf("%s pid %d is wanting ...\n", prefix, pid);
}

static void acquired(char *method, char *prefix)
{
	pid_t pid;

	pid = getpid();
	if (!prefix)
		printf("pid %d acquired %s\n", pid, method);
	else
		printf("%s pid %d acquired %s\n", prefix, pid, method);
}

static void releases(char *prefix)
{
	pid_t pid;

	pid = getpid();
	if (!prefix)
		printf("pid %d releases\n", pid);
	else
		printf("%s pid %d releases\n", prefix, pid);
}

/* ---------------------------------------------------------------------- */

static int lock_by_flock(int fd, char *prefix)
{
	int err;

	wanting(prefix);
	err = flock(fd, LOCK_EX);
	if (!err)
		acquired("flock", prefix);
	return err;
}

static int unlock_by_flock(int fd, char *prefix)
{
	releases(prefix);
	return flock(fd, LOCK_UN);
}

static struct lock_ops flock_ops = {
	.lock		= lock_by_flock,
	.unlock		= unlock_by_flock
};

/* ---------------------------------------------------------------------- */

static int lock_by_posix(int fd, char *prefix)
{
	int err;
	struct flock lk = {
		.l_type		= F_WRLCK,
		.l_whence	= SEEK_SET,
		.l_start	= 0,
		.l_len		= 0 /* whole file */
	};

	wanting(prefix);
	err = fcntl(fd, F_SETLKW, &lk);
	if (!err)
		acquired("posix", prefix);
	return err;
}

static int unlock_by_posix(int fd, char *prefix)
{
	struct flock lk = {
		.l_type		= F_UNLCK,
		.l_whence	= SEEK_SET,
		.l_start	= 0,
		.l_len		= 0 /* whole file */
	};

	releases(prefix);
	return fcntl(fd, F_SETLKW, &lk);
}

static struct lock_ops posix_ops = {
	.lock		= lock_by_posix,
	.unlock		= unlock_by_posix
};

/* ---------------------------------------------------------------------- */

static int lock_by_ofd(int fd, char *prefix)
{
	int err;
	struct flock lk = {
		.l_type		= F_WRLCK,
		.l_whence	= SEEK_SET,
		.l_start	= 0,
		.l_len		= 0, /* whole file */
		.l_pid		= 0
	};

	wanting(prefix);
	err = fcntl(fd, F_OFD_SETLKW, &lk);
	if (!err)
		acquired("ofd", prefix);
	return err;
}

static int unlock_by_ofd(int fd, char *prefix)
{
	struct flock lk = {
		.l_type		= F_UNLCK,
		.l_whence	= SEEK_SET,
		.l_start	= 0,
		.l_len		= 0, /* whole file */
		.l_pid		= 0
	};

	releases(prefix);
	return fcntl(fd, F_OFD_SETLKW, &lk);
}

static struct lock_ops ofd_ops = {
	.lock		= lock_by_ofd,
	.unlock		= unlock_by_ofd
};

/* ---------------------------------------------------------------------- */

struct lock_ops *lock_ops(char *str)
{
	if (!strcmp(str, "bsd")
	    || !strcmp(str, "flock"))
		return &flock_ops;
	else if (!strcmp(str, "posix")
		 || !strcmp(str, "setlkw")
		 || !strcmp(str, "fcntl"))
		return &posix_ops;
	else if (!strcmp(str, "ofd")
		 || !strcmp(str, "ofd_setlkw"))
		return &ofd_ops;

	assert(0);
	return NULL;
}

/* ---------------------------------------------------------------------- */

int open_and_pat(char *fname, int flags, char pat[], int n)
{
	int fd, err, len;
	struct stat st;

	fd = open(fname, flags);
	assert(fd >= 0);
	err = fstat(fd, &st);
	assert(!err);
	len = snprintf(pat, n, "%02x:%02x:%lu",
		       major(st.st_dev), minor(st.st_dev),
		       (unsigned long)st.st_ino);
	assert(len < n);
	//printf("%s is %s\n", fname, pat);
	return fd;
}

void grep_lock(char pat[], const char func[], int line)
{
	int found;
	char a[128];
	FILE *fp;

	fp = fopen("/proc/locks", "r");
	assert(fp);
	found = 0;
	while (fgets(a, sizeof(a), fp))
		if (strstr(a, pat)) {
			found++;
			printf("%s:%d: %s", func, line, a);
		}
	fclose(fp);
	if (!found)
		printf("%s:%d: pid %d has no locks on %s\n",
		       func, line, getpid(), pat);
}
