/*
 * Copyright (C) 2014 SENJU Jiro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _GNU_SOURCE
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <assert.h>
#include <fcntl.h>
#include <mcheck.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>


#if __GLIBC_PREREQ(2, 14)
#else
#define O_PATH		010000000	/* linux/include/uapi/asm-generic/fcntl.h */

#define MAX_HANDLE_SZ 128	/* linux/include/linux/exportfs.h */

struct file_handle {
	unsigned int handle_bytes;
	int handle_type;
	unsigned char f_handle[0];
};

static inline
int name_to_handle_at(int dfd, const char *name, struct file_handle *handle,
		      int *mnt_id, int flags)
{
	return syscall(SYS_name_to_handle_at, dfd, name, handle, mnt_id, flags);
}

static inline
int open_by_handle_at(int mountdirfd, struct file_handle *handle, int flags)
{
	return syscall(SYS_open_by_handle_at, mountdirfd, handle, flags);
}
#endif

/* ---------------------------------------------------------------------- */

struct pipedata {
	int mnt_id;
	struct file_handle fh;
};

/* ---------------------------------------------------------------------- */

int child_handle(char *file, int outfd)
{
	int err;
	ssize_t ssz;
	struct file_handle *fh;
	struct pipedata pd;

	fh = malloc(sizeof(*fh) + MAX_HANDLE_SZ);
	assert(fh);
	fh->handle_bytes = MAX_HANDLE_SZ;
	err = name_to_handle_at(AT_FDCWD, file, fh, &pd.mnt_id, 0);
	assert(!err);

	pd.fh = *fh;
	ssz = write(outfd, &pd, sizeof(pd));
	assert(ssz == sizeof(pd));
	ssz = write(outfd, fh->f_handle, fh->handle_bytes);
	assert(ssz == fh->handle_bytes);
	printf("[%d] fhandle is generated\n", getpid());

	free(fh);

	return 0;
}

/* ---------------------------------------------------------------------- */

//#define TryO_PATH

int mntfd(int mnt_id)
{
	int dirfd, id, n;
	FILE *fp;
	char *mntpnt, *p, line[PATH_MAX];

	mntpnt = NULL;
	fp = fopen("/proc/self/mountinfo", "r");
	assert(fp);
	do {
		free(mntpnt);
		p = fgets(line, sizeof(line), fp);
		assert(p == line);
		n = sscanf(line, "%d %*d %*s %*s %as\n", &id, &mntpnt);
		assert(n == 2);
	} while (id != mnt_id);
	fclose(fp);
	assert(id == mnt_id);

#ifdef TryO_PATH
	int err;

	dirfd = open(mntpnt, O_RDONLY | O_PATH);
	assert(dirfd >= 0);
	err = fchdir(dirfd);
	assert(!err);
	close(dirfd);
	dirfd = AT_FDCWD;
#else
	dirfd = open(mntpnt, O_RDONLY);
	assert(dirfd >= 0);
#endif

	free(mntpnt);

	return dirfd;
}

int child_open(int infd)
{
	int err, fd, dirfd;
	char *filedata;
	struct stat st;
	ssize_t ssz;
	struct pipedata pd;

	mtrace();
	ssz = read(infd, &pd, sizeof(pd));
	assert(ssz == sizeof(pd));
	ssz = read(infd, pd.fh.f_handle, pd.fh.handle_bytes);
	assert(ssz == pd.fh.handle_bytes);
	printf("[%d] fhandle is read\n", getpid());

	dirfd = mntfd(pd.mnt_id);
	fd = open_by_handle_at(dirfd, &pd.fh, O_RDONLY);
	assert(fd >= 0);
	err = fstat(fd, &st);
	assert(!err);
	filedata = malloc(st.st_size + 1);
	assert(filedata);
	ssz = read(fd, filedata, st.st_size);
	assert(ssz == st.st_size);
	filedata[st.st_size] = '\0';
	puts(filedata);
	free(filedata);
	if (dirfd != AT_FDCWD)
		close(dirfd);

	muntrace();
	return 0;
}

/* ---------------------------------------------------------------------- */

int main(int argc, char *argv[])
{
	int err, fd, status, pfd[2], i;
	ssize_t ssz;
	pid_t pid[2], waited __attribute__((unused));
	char msg[] = "hello, world";

	fd = open(argv[1], O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
	assert(fd >= 0);
	ssz = write(fd, msg, sizeof(msg) - 1);
	assert(ssz == sizeof(msg) - 1);
	err = close(fd);
	assert(!err);

	err = pipe(pfd);
	assert(!err);
	pid[0] = fork();
	assert(pid[0] >= 0);
	if (!pid[0]) {
		err = child_handle(argv[1], pfd[1]);
		exit(err ? EXIT_SUCCESS : EXIT_FAILURE);
	}
	pid[1] = fork();
	assert(pid[1] >= 0);
	if (!pid[1]) {
		err = child_open(pfd[0]);
		exit(err ? EXIT_SUCCESS : EXIT_FAILURE);
	}
	for (i = 0; i < 2; i++) {
		waited = waitpid(pid[0], &status, 0);
		assert(WIFEXITED(status));
	}

	return 0;
}
